
/*-----------------------------------------------------------------------------
|
|   FreePort - This utility provides a simple mechanism to communicate to the
|              NSLink device driver that it should request the physical device
|              to close the physical ports which are associated with the
|              specified logical ports.
|
|              Additionally, it provides a simple example on how other
|              applications could implement this capability themselves.
|
|
|   Copyright (c) 2001-2006 Comtrol Corporation All Rights Reserved
|
|-----------------------------------------------------------------------------*/

#define STRICT

#include <windows.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) )

#define METHOD_BUFFERED                 0
#define FILE_ANY_ACCESS                 0
#define FILE_DEVICE_SERIAL_PORT         0x0000001b

#define IOCTL_RELEASE_NSLINK_PORT \
      CTL_CODE(FILE_DEVICE_SERIAL_PORT, 0x815,METHOD_BUFFERED,FILE_ANY_ACCESS)

typedef struct _PortNode  *PPortNode;

typedef struct _PortNode
{
    int       Num;
    PPortNode Next;
} PortNode; 

int parse_options (int argc, char *argv[]);

PPortNode PortList;

// The following few lines are needed to support NS-Link 64-bit driver
//
typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, PBOOL);

LPFN_ISWOW64PROCESS fnIsWow64Process = 
 (LPFN_ISWOW64PROCESS)GetProcAddress(GetModuleHandle("kernel32"), "IsWow64Process");

// Are we running on an 64-bit OS? 
BOOL IsWow64()
{
    BOOL bIsWow64 = FALSE;
 
    if (NULL != fnIsWow64Process)
    {
        if (!fnIsWow64Process(GetCurrentProcess(),&bIsWow64))
        {
            return TRUE;
        }
    }
    return bIsWow64;
}

/*-----------------------------------------------------------------------------
|
|  main - Begins by trying to open a communications channel to the driver.  It
|         then builds a list of ports from the command line specification.  If
|         a proper specification has occurred, the list is then processed, one
|         port at a time, by prepending "COM" to the port number to obtain the
|         full specification which the driver expects.  The call is then made
|         to pass the port name to the driver with the I/O Control code which
|         instructs the driver to request the device upon which the physical
|         port resides to unlock that port for someone else to acquire.
|         Finally, all allocated resources are freed and control is returned.
|
|         If at any time a catastrophic error occurs, notification of the error
|         occurring is made, all allocated resources are returned and control
|         returns.
|
|-----------------------------------------------------------------------------*/

int main (int argc, char *argv[])
{
    int stat; 

    PPortNode CurrentPort;

    TCHAR *DriverName;

    HANDLE hDevice;

    char  *PortNum,
          *InBuffer,
          *OutBuffer;

    DWORD BytesReturned;

    DWORD AccessMode = GENERIC_READ | GENERIC_WRITE;

    DWORD ShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE;

    PortNum   = (char *) malloc (3);
    InBuffer  = (char *) malloc (6);
    OutBuffer = (char *) malloc (6);

	// Identify the right driver name based on the OS 
	// and Processer architecture
	//
	if (IsWow64())
	{
		SYSTEM_INFO	sysInfo;
		GetSystemInfo(&sysInfo);

		if (sysInfo.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_IA64)
			DriverName = TEXT("\\\\.\\rpshsiia64");	// IA64 driver
		else
			DriverName = TEXT("\\\\.\\rpshsix64");	// X64 driver
	}
	else	
		DriverName = TEXT("\\\\.\\rpshsi");	// X32 driver

    //
    // Establish connection to device driver and save handle for channel
    //

    hDevice = CreateFile (DriverName,
                          AccessMode,
                          ShareMode,
                          0,
                          OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL);

    if ( (hDevice == INVALID_HANDLE_VALUE) || (hDevice == NULL) )
    {
        printf ("Unable to communicate with driver '%s'\n", DriverName);
        CloseHandle (hDevice);
        free (PortNum);
        free (InBuffer);
        free (OutBuffer);
        return 1;
    };

    //
    // Obtain list of COM ports upon which device is to force a release.
    //

    if (parse_options (argc, argv))
    {
        CloseHandle (hDevice);
        free (PortNum);
        free (InBuffer);
        free (OutBuffer);
        return 1;
    };

    //
    // Cycle through the list and attempt to have the driver instruct the
    // physical device to unlock the physical port.
    //

    CurrentPort = PortList;

    while (CurrentPort)
    {
        InBuffer = strcpy (InBuffer, "COM");

        itoa (CurrentPort->Num, PortNum, 10);

        InBuffer = strcat (InBuffer, PortNum);

        stat = DeviceIoControl (hDevice,
                                IOCTL_RELEASE_NSLINK_PORT,
                                InBuffer,
                                strlen (InBuffer),
                                OutBuffer,
                                6,
                                &BytesReturned,
                                0);

        if (stat != TRUE)
        {
            int stop = 0;

            stat = GetLastError();

            switch (stat)
            {
                case ERROR_UNEXP_NET_ERR:
                    printf ("Unable to communicate with device for port %s\n",
                            InBuffer);
                    break;

                case ERROR_INSUFFICIENT_BUFFER:
                    stop = 1;
                    break;

                case ERROR_INVALID_HANDLE:
                    printf ("Port %s is not an NSLink Port\n", InBuffer);
                    break;

                case ERROR_NO_SYSTEM_RESOURCES:
                    stop = 1;
                    break;
            };

            if (stop)
            {
                CloseHandle (hDevice);
                free (PortNum);
                free (InBuffer);
                free (OutBuffer);
                return 1;
            };
        };

        CurrentPort = CurrentPort->Next;
    };

    CloseHandle (hDevice);
    free (PortNum);
    free (InBuffer);
    free (OutBuffer);
    return 0;
};

/*-----------------------------------------------------------------------------
|
|  DisplayHelp - Displays the correct command syntax which should have been
|                issued.
|
|-----------------------------------------------------------------------------*/

void display_help ()
{

    printf ("\n\nInvalid usage:  FreePort <comspec>\n");
    printf ("\n    where:\n");

    printf ("        <comspec> - COM port specification using one or more\n");
    printf ("                    of the following forms :'\n\n");

    printf ("          x   - number of a single COM port (i.e.  3 for COM3)\n");

    printf ("          x-z - a range of COM ports (i.e. COMx, COMy, COMz)\n\n");
    printf ("                    if more than one form is specified, they\n");
    printf ("                    must be separated by a comma (i.e. 5,7-10,9)\n");
};


/*-----------------------------------------------------------------------------
|
|  parse_options - Parses through the specified command line parameters and
|                  builds a linked list of the port numbers specified there.
|
|-----------------------------------------------------------------------------*/


int parse_options (int argc, char *argv[])
{
    int dex,
        length,
        LowerPort,
        UpperPort;

    PPortNode PNode,
              CurNode;

    char *ComSpec = argv[1];

    if (argc < 2)
    {
        display_help();
        return 1;
    };

    dex = 0;
    CurNode = NULL;
    PortList = NULL;
    length = (int) strlen (ComSpec);

    while (dex < length)
    {
        if (!isdigit ((int) ComSpec[dex]))
        {
            display_help();
            return 1;
        };

        LowerPort = 0;
        UpperPort = 0;

        while  (isdigit ( (int) ComSpec[dex]))
        {
            if (LowerPort)
                LowerPort *=10;

            LowerPort +=  (ComSpec[dex++]) - '0';
        };

        if (!memcmp ("-", &ComSpec[dex], 1))
        {
            dex++;
            while  (isdigit ( (int) ComSpec[dex]))
            {
                if (UpperPort)
                    UpperPort *=10;

                UpperPort +=  (ComSpec[dex++]) - '0';
            };
        };

        if (dex < length)
        {
            if (!memcmp (",", &ComSpec[dex], 1))
            {
                dex++;
            }
            else
            {
                display_help();
                return 1;
            };
        };

        do
        {
            PNode = (PortNode *) malloc (sizeof (PortNode));

            PNode->Num = LowerPort++;
            PNode->Next    = NULL;

            if (CurNode)
            {
                CurNode->Next = PNode;
            }
            else
            {
               PortList = PNode;
            };

            CurNode = PNode;

        } while (LowerPort < UpperPort + 1);

    };

    return 0;
};
